import tempfile

from streamduo.api.key import KeyController
from streamduo.models.batch_data import BatchData
from streamduo.models.batch_init_request import BatchInitRequest
import hashlib
import os
import base64

from streamduo.models.schema import SchemaType

from streamduo.models.schema import FileType
from streamduo.validators.json_schema import JsonValidator
from streamduo.validators.great_expectations_schema import GreatExepectationsValidator


class BatchController:
    """
    Provides methods for interacting with the `/batch` endpoints
    """
    BUFFER_SIZE = 1024 * 1024 * 5  # 5 MB

    def __init__(self, client):
        self.client = client

    @staticmethod
    def construct_batch_init_request(file_path, file_name_override=None):
        """
        constructs the payload for the initiate batch file upload request
        :param file_path: (STRING) path to file for upload.
        :param file_name_override: (STRING) optional override for file name
        :return: (BatchInitRequest) object.
        """
        # construct request
        batch_init_request = BatchInitRequest()
        part_number = 1
        # break down chunks
        with open(file_path, 'rb') as out_file:
            full_file_md5 = hashlib.md5()
            while True:
                data = out_file.read(BatchController.BUFFER_SIZE)
                if not data:
                    break
                # update full file Hash
                full_file_md5.update(data)
                # add part hash as b64 to batch init req
                batch_init_request.hashes[part_number] = base64.b64encode(
                    hashlib.md5(data).digest()).decode()
                part_number = part_number + 1

        batch_init_request.hashValue = base64.b64encode(full_file_md5.digest()).decode()
        if file_name_override:
            file_name_metadata = file_name_override
        else:
            file_name_metadata = file_path
        batch_init_request.fileName = os.path.basename(file_name_metadata)
        file_name, file_extension = os.path.splitext(file_name_metadata)
        file_extension = file_extension.strip('.').upper()
        if file_extension in FileType._member_names_:
            batch_init_request.fileType = FileType[file_extension].value
        batch_init_request.totalParts = part_number
        return batch_init_request

    @staticmethod
    def get_part_binary(file_path, part_number):
        """
        returns the bytes for a particular "part number" of a file.
        :param file_path: (STRING) path to file for upload.
        :param part_number: (INT) part number to return
        :return: (BYTES) bytes of data corresponding to the file part.
        """
        part_counter = 1
        with open(file_path, 'rb') as out_file:
            while part_counter <= int(part_number):
                data = out_file.read(BatchController.BUFFER_SIZE)
                # EOF
                if not data:
                    break
                part_counter = part_counter + 1
            return data

    def send_batch_init(self, stream_id, file_path, file_path_override=None):
        """
        Sends a Batch initiation request to the API.
        :param stream_id: (STRING) Stream ID file will be uploaded to.
        :param file_path: (STRING) Path to file being uploaded.
        :param file_path_override: (STRING) optional overrride for file path
        :return:
        """
        request_object = BatchController.construct_batch_init_request(file_path=file_path,
                                                                      file_name_override=file_path_override)
        return self.client.call_api('POST',
                                    f"/stream/{stream_id}/batch/init",
                                    body=request_object.to_json())

    def send_batch_part(self, batch_data, part_number, binary_payload):
        file_payload = {'file': binary_payload}
        return self.client.call_api('POST',
                                    f"/stream/{batch_data.streamId}/batch/{batch_data.batchId}/upload-part/{part_number}",
                                    files=file_payload)

    def send_file(self, stream_id, file_path):
        ## init req
        batch_data = BatchData(**self.send_batch_init(stream_id=stream_id,
                                                      file_path=file_path).json())

        ## CSV w. JSON Schema
        if batch_data.requires_validation() \
                and batch_data.get_file_type() == FileType.CSV \
                and batch_data.get_stream_schema().get_schema_type() == SchemaType.JSON:
            ## validate
            val = JsonValidator()
            val.set_schema(batch_data.get_stream_schema().get_schema())
            val.validate_csv(file_path)

        ## CSV w. GE Schema
        if batch_data.requires_validation() \
                and batch_data.get_file_type() == FileType.CSV \
                and batch_data.get_stream_schema().get_schema_type() == SchemaType.GREAT_EXPECTATIONS:
            ## validate
            val = GreatExepectationsValidator()
            val.set_schema(batch_data.get_stream_schema().get_schema())
            val.validate_csv(file_path)

        ## Encryption
        if batch_data.requires_encryption():
            with tempfile.NamedTemporaryFile() as tmp:
                KeyController.encrypt_file(key_string=batch_data.get_public_key_value(), source_file_path=file_path,
                                           destination_file_path=tmp.name)
                ## recreate batch_data with encrypted file hashes
                batch_data = BatchData(**self.send_batch_init(stream_id=stream_id,
                                                              file_path=tmp.name,
                                                              file_path_override=file_path).json())
                return self.send_chunks(batch_data=batch_data, file_path=tmp.name)

        else:
            return self.send_chunks(batch_data=batch_data, file_path=file_path)

    def send_chunks(self, batch_data, file_path) -> BatchData:
        # send via loop chunks
        with open(file_path, 'rb') as out_file:
            while len(batch_data.outstandingParts.keys()) > 0:
                part_number = next(iter(batch_data.outstandingParts.items()))[0]
                data = BatchController.get_part_binary(file_path=file_path, part_number=part_number)
                # send data
                batch_data = BatchData(**self.send_batch_part(batch_data=batch_data,
                                                              part_number=part_number, binary_payload=data).json())
        return batch_data

    def get_part(self, stream_id, batch_id, part_number) -> bytes:

        resp = self.client.call_api('GET',
                                    f"/stream/{stream_id}/batch/{batch_id}/get-part/{part_number}")
        return resp.content

    def get_batch(self, stream_id: str, batch_id: str, destination_filepath: str, decryption_key: str = None):
        ## get batchdata
        batch_data = self.get_batch_metadata(stream_id=stream_id, batch_id=batch_id)
        outstanding_parts = batch_data.hashes.copy()
        with tempfile.TemporaryDirectory() as td:
            while len(outstanding_parts.keys()) > 0:
                part_number = next(iter(outstanding_parts.items()))[0]
                file_name = os.path.join(td, f"{stream_id}-{batch_id}-{str(part_number)}.part")
                part_data = self.get_part(stream_id=stream_id, batch_id=batch_id, part_number=part_number)
                ## validate part download
                if outstanding_parts[part_number] != base64.b64encode(hashlib.md5(part_data).digest()).decode():
                    raise ValueError
                with open(file_name, 'wb') as fh:
                    fh.write(part_data)
                del outstanding_parts[part_number]
            ## Compile final file
            final_file_tmp = os.path.join(td, f"{stream_id}-{batch_id}.final")
            with open(final_file_tmp, 'wb') as final_file:
                for pn in batch_data.hashes.keys():
                    with open(os.path.join(td, f"{stream_id}-{batch_id}-{str(pn)}.part"), 'rb') as part_file:
                        final_file.write(part_file.read())
            if batch_data.requires_encryption():
                KeyController.decrypt_file(source_file_path=final_file_tmp,
                                           destination_file_path=destination_filepath,
                                           key_string=decryption_key)
            else:
                os.rename(final_file_tmp, destination_filepath)

    def get_batch_metadata(self, stream_id: str, batch_id: str) -> BatchData:
        resp = self.client.call_api('GET', f"/stream/{stream_id}/batch/{batch_id}")
        return BatchData(**resp.json())



from nacl.encoding import Base64Encoder
from nacl.public import PrivateKey, PublicKey, SealedBox
from requests import Response

from streamduo.models.key_pair import KeyPair, KeyEncoding, KeyAlgorithm


class KeyController:
    """
    Provides methods for interacting with the `/key` endpoints
    """

    def __init__(self, client):
        self.client = client

    @staticmethod
    def create_key_local() -> tuple[KeyPair, str]:
        """
        Creates a new pub/priv keypair locally
        :return: (tuple) Public Key Object, Private Key String
        """
        ## create a secret Key
        sk = PrivateKey.generate()
        pk = KeyPair(publicKey=sk.public_key.encode(encoder=Base64Encoder).decode('utf8'),
                     publicKeyAlgorithm=KeyAlgorithm.CURVE25519.value,
                     publicKeyEncoding=KeyEncoding.BASE64.value)
        return  pk, sk.encode(encoder=Base64Encoder).decode('utf8')

    @staticmethod
    def get_public_key(key_string: str) -> PublicKey:
        """
        Converts a public Key String to a PublicKey Object
        :param key_string: Base64Encoded String of Public Key
        :return: PublicKey Object
        """
        return PublicKey(public_key=key_string.encode('utf-8'), encoder=Base64Encoder)

    @staticmethod
    def get_private_key(key_string: str) -> PrivateKey:
        """
        Converts a Private Key String to a PrivateKey Object
        :param key_string: Base64Encoded String of Private Key
        :return: PrivateKey Object
        """
        return PrivateKey(private_key=key_string.encode('utf-8'), encoder=Base64Encoder)

    @staticmethod
    def encrypt_file(key_string: str, source_file_path: str, destination_file_path: str):
        """
        Static Method for encrypting a file with a given public key String
        :param key_string: Base64Encoded String of public key
        :param source_file_path: filepath for source data
        :param destination_file_path: filepath for destination
        :return: None
        """
        sb = SealedBox(KeyController.get_public_key(key_string))
        with open(source_file_path, 'rb') as in_file:
            encrypted_data = sb.encrypt(in_file.read())
        with open(destination_file_path, 'wb') as out_file:
            out_file.write(encrypted_data)

    @staticmethod
    def decrypt_file(key_string: str, source_file_path: str, destination_file_path: str):
        """
         Static Method for decrypting a file with a given private key String
        :param key_string: Base64Encoded String of private key
        :param source_file_path:  filepath for source encrypted data
        :param destination_file_path:  filepath for destination decrypted data
        :return: None
        """
        unseal_box = SealedBox(KeyController.get_private_key(key_string))
        with open(source_file_path, 'rb') as in_file:
            out = unseal_box.decrypt(in_file.read())
        with open(destination_file_path, 'wb') as out_file:
            out_file.write(out)

    def upload_key(self, stream_id: str, key: str) -> Response:
        """
        Upload a locally generated public key to StreamDuo stream
        :param stream_id: ID of Stream to upload key to.
        :param key: Encoded string of public key.
        :return: requests.response object
        """
        return self.client.call_api('POST',
                                    f"/stream/{stream_id}/keys/",
                                    body=key.to_json())

    def set_active_key(self, stream_id: str, key_id: str) -> Response:
        """
        Set a key active on a streamDuo Stream
        :param stream_id: ID of Stream to set key active on
        :param key_id: ID of Key to set active
        :return: requests.response object
        """
        return self.client.call_api('POST',
                                    f"/stream/{stream_id}/keys/active",
                                    body={'keyId': key_id})

    def create_key_server(self, stream_id: str, key_request: dict) -> Response:
        """
        Server Side Key Creation
        :param stream_id: ID of Stream to create key on
        :param key_request: dict of request params
        :return:  requests.response object
        """
        return self.client.call_api('POST',
                                    f"/stream/{stream_id}/keys/generate",
                                    body=key_request)



import filecmp
import json
import os
import pprint
from unittest import TestCase

from requests import HTTPError

from streamduo.client import Client
from streamduo.models.batch_data import BatchData
from streamduo.api.batch import BatchController
from streamduo.models.key_pair import KeyAlgorithm, KeyEncoding
from streamduo.models.schema import SchemaType
from streamduo.api.key import KeyController


class TestBatch(TestCase):

    def setUp(self) -> None:
        self.api_client = Client(os.getenv('AUTH_CLIENT_ID'), os.getenv('AUTH_CLIENT_SECRET'))

    def test_init_batch(self):
        display_name = 'batch_test_stream'
        stream_controller = self.api_client.get_stream_controller()
        new_stream_id = stream_controller.create_stream(display_name).json()['streamId']

        batch_controller = self.api_client.get_batch_controller()
        FILE_SIZE = 1024 * 1024 * 100  # 1GB
        big_file = "bigfile.dat"
        # multiply file
        with open(big_file, 'wb') as out_file:
            out_file.write(os.urandom(FILE_SIZE))

        resp = batch_controller.send_batch_init(stream_id=new_stream_id, file_path=big_file)
        batch_data = BatchData(**resp.json())
        assert len(batch_data.outstandingParts.keys()) == 20
        assert batch_data.streamSchema == None
        assert batch_data.requires_validation() == False

        ## Add schema to stream
        schema_controller = self.api_client.get_schema_controller()
        with open("car_schema.json", 'r') as file:
            car_schema = json.load(file)
        create_schema_response1 = schema_controller.create_schema(stream_id=new_stream_id,
                                                                  schema=car_schema, schema_type=SchemaType.JSON)
        new_schema_id = create_schema_response1.json()['schemaId']

        ## Set Active Schema
        set_active_response = schema_controller.set_active_schema(stream_id=new_stream_id, schema_id=new_schema_id)

        ## Create new Req
        resp = batch_controller.send_batch_init(stream_id=new_stream_id, file_path=big_file)
        batch_data = BatchData(**resp.json())
        assert len(batch_data.outstandingParts.keys()) == 20
        assert batch_data.requires_validation() == True
        assert batch_data.get_stream_schema().get_schema()['$id'] == 'https://example.com/object1644429127.json'

        ## Cleanup
        stream_controller.delete_stream(new_stream_id)

    def test_send_part(self):
        display_name = 'batch_test_stream'
        stream_controller = self.api_client.get_stream_controller()
        new_stream_id = stream_controller.create_stream(display_name).json()['streamId']

        batch_controller = self.api_client.get_batch_controller()
        FILE_SIZE = 1024 * 1024 * 100  # 1GB
        big_file = "bigfile.dat"
        # multiply file
        with open(big_file, 'wb') as out_file:
            out_file.write(os.urandom(FILE_SIZE))

        resp = batch_controller.send_batch_init(stream_id=new_stream_id, file_path=big_file)
        batch_data = BatchData(**resp.json())
        part_bin = BatchController.get_part_binary(file_path=big_file, part_number=5)
        ## Send part 5
        batch_data_resp = batch_controller.send_batch_part(batch_data=batch_data, part_number=5,
                                                           binary_payload=part_bin)
        batch_data_2 = BatchData(**batch_data_resp.json())
        pprint.pprint(batch_data_2.to_json())
        assert '5' not in batch_data_2.outstandingParts.keys()

        ## Cleanup
        stream_controller.delete_stream(new_stream_id)

    def test_full_file(self):
        display_name = 'batch_test_stream_full'
        stream_controller = self.api_client.get_stream_controller()
        new_stream_id = stream_controller.create_stream(display_name).json()['streamId']

        batch_controller = self.api_client.get_batch_controller()
        FILE_SIZE = 1024 * 1024 * 20  # 20MB
        big_file = "bigfile.dat"
        # multiply file
        with open(big_file, 'wb') as out_file:
            out_file.write(os.urandom(FILE_SIZE))

        batch_data = batch_controller.send_file(stream_id=new_stream_id, file_path=big_file)
        assert len(batch_data.outstandingParts.keys()) == 0
        ## Cleanup
        stream_controller.delete_stream(new_stream_id)

    def test_full_file_schema(self):
        display_name = 'batch_test_stream_full_schema'
        stream_controller = self.api_client.get_stream_controller()
        new_stream_id = stream_controller.create_stream(display_name).json()['streamId']

        ## Add schema to stream
        schema_controller = self.api_client.get_schema_controller()
        with open("car_schema.json", 'r') as file:
            car_schema = json.load(file)
        create_schema_response1 = schema_controller.create_schema(stream_id=new_stream_id,
                                                                  schema=car_schema, schema_type=SchemaType.JSON)
        new_schema_id = create_schema_response1.json()['schemaId']

        ## Set Active Schema
        schema_controller.set_active_schema(stream_id=new_stream_id, schema_id=new_schema_id)

        ## Send File
        batch_controller = self.api_client.get_batch_controller()
        batch_data = batch_controller.send_file(stream_id=new_stream_id, file_path="../test_data/car_sales.csv")
        assert len(batch_data.outstandingParts.keys()) == 0

        ## Cleanup
        stream_controller.delete_stream(new_stream_id)

    def test_full_file_schema_enc(self):
        display_name = 'batch_test_stream_full_schema_enc'
        stream_controller = self.api_client.get_stream_controller()
        new_stream_id = stream_controller.create_stream(display_name).json()['streamId']

        ## Add schema to stream
        schema_controller = self.api_client.get_schema_controller()
        with open("car_schema.json", 'r') as file:
            car_schema = json.load(file)
        create_schema_response1 = schema_controller.create_schema(stream_id=new_stream_id,
                                                                  schema=car_schema, schema_type=SchemaType.JSON)
        new_schema_id = create_schema_response1.json()['schemaId']

        ## Set Active Schema
        schema_controller.set_active_schema(stream_id=new_stream_id, schema_id=new_schema_id)

        ## Create Key
        key_controller = self.api_client.get_key_controller()
        key_request = {'publicKeyAlgorithm': KeyAlgorithm.CURVE25519.value,
                       'publicKeyEncoding': KeyEncoding.BASE64.value,
                       'publicKeyDescription': "test key integration"}
        new_key_resp = key_controller.create_key_server(stream_id=new_stream_id, key_request=key_request)
        priv = new_key_resp.json()['privateKeyValue']
        pub = new_key_resp.json()['publicKey']['publicKeyValue']
        key_id = new_key_resp.json()['publicKey']['publicKeyId']

        ## set Active
        key_controller.set_active_key(stream_id=new_stream_id, key_id=key_id)

        ## Send File
        original_filepath = "../test_data/car_sales.csv"
        batch_controller = self.api_client.get_batch_controller()
        batch_data = batch_controller.send_file(stream_id=new_stream_id, file_path=original_filepath)
        assert len(batch_data.outstandingParts.keys()) == 0
        print(f"streamId: {new_stream_id}, Batch ID: {batch_data.batchId}, priv key: {priv}")

        dest_filepath = 'out_dec.csv'
        batch_controller.get_batch(stream_id=new_stream_id,
                                   batch_id=batch_data.batchId,
                                   destination_filepath=dest_filepath,
                                   decryption_key=priv)

        assert filecmp.cmp(original_filepath, dest_filepath)
        ## Cleanup
        stream_controller.delete_stream(new_stream_id)
        os.remove(dest_filepath)